/*
 * Copyright (c) 2008 Jiri Svoboda
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup debug
 * @{
 */
/** @file
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <udebug.h>
#include <kernel/arch/context_offset.h>

#include "../../../cons.h"
#include "../../../main.h"
#include "../../../breakpoint.h"
#include "../../../include/arch.h"

#define OPCODE_BREAK		0x0000000d

static unsigned buffer[1024];

int arch_breakpoint_set(breakpoint_t *b)
{
	uint32_t brkp;
	int rc;

	rc = udebug_mem_read(app_phone, &b->arch.back, b->addr,
	    sizeof(b->arch.back));
	if (rc < 0) return rc;

	cons_printf("udebug_mem_read() -> %d\n", rc);
	brkp = OPCODE_BREAK;
	rc = udebug_mem_write(app_phone, &brkp, b->addr, sizeof(brkp));
	cons_printf("udebug_mem_write() -> %d\n", rc);
	if (rc < 0) return rc;

	return 0;
}

int arch_breakpoint_remove(breakpoint_t *b)
{
	int rc;

	if (b->active) {
		rc = udebug_mem_write(app_phone, &b->arch.back, b->addr + 4, 1);
		if (rc < 0) {
			cons_printf("error writing memory\n");
			return rc;
		}
		active_bkpt = NULL;
	} else {
	        rc = udebug_mem_write(app_phone, &b->arch.back, b->addr, 1);
		if (rc < 0) {
			cons_printf("error writing memory\n");
			return rc;
		}
	}

	return 0;

}

void arch_event_breakpoint(thash_t thread_hash)
{
	breakpoint_t *b;
	int rc;
	uint32_t epc;
	int brk_addr;
	uint32_t brkp;

	brkp = OPCODE_BREAK;

	rc = udebug_regs_read(app_phone, thread_hash, buffer);
	cons_printf("udebug_regs_read -> %d\n", rc);
	epc = buffer[EOFFSET_EPC/sizeof(unsigned)];
	cons_printf("EPC was 0x%08x\n", epc);
	brk_addr = epc;

	b = breakpoint_find_by_addr(brk_addr);
	if (b != NULL) {
	        rc = udebug_mem_write(app_phone, &b->arch.back, brk_addr, 4);
		cons_printf("udebug_mem_write(phone, 0x%x, 0x%02x, 1) -> %d\n", brk_addr, b->arch.back, rc);
		rc = udebug_mem_read(app_phone, &b->arch.back, brk_addr + 4, 4);
	        rc = udebug_mem_write(app_phone, &brkp, brk_addr + 4, 4);
		active_bkpt = b;

		breakpoint_hit(b);
	}

	b = breakpoint_find_by_addr(brk_addr - 4);
	if (b != NULL && b->active) {
		cons_printf("restoring breakpoint %d\n", b->id);
	        rc = udebug_mem_write(app_phone, &b->arch.back, brk_addr, 4);
		rc = udebug_mem_read(app_phone, &b->arch.back, brk_addr - 4, 4);
	        rc = udebug_mem_write(app_phone, &brkp, brk_addr - 4, 4);
		active_bkpt = NULL;
		return;	
	}

	cons_printf("Unrecognized breakpoint at 0x%lx\n", brk_addr);
}

void arch_event_trap(dthread_t *dt)
{
	/* Unused */
	(void)dt;
}

void arch_dump_regs(thash_t thash)
{
}

void arch_set_singlestep(dthread_t *dt, int enable)
{
}

/** @}
 */
